# xDevs.com Calibration app for Keithley 2304, v0.1
# https://xdevs.com/fix/kei2304a
# https://xdevs.com/fix/hp3458a
import sys
import time
import Gpib
import signal
import numbers
from math import log10

dmm_val = 1.0
dmm_temp = 36.6

class Timeout():
  """Timeout class using ALARM signal"""
  class Timeout(Exception): pass

  def __init__(self, sec):
    self.sec = sec

  def __enter__(self):
    signal.signal(signal.SIGALRM, self.raise_timeout)
    signal.alarm(self.sec)

  def __exit__(self, *args):
    signal.alarm(0) # disable alarm

  def raise_timeout(self, *args):
    raise Timeout.Timeout()

class dmm():
    global dmm_val 
    data = ""
    temp = 36.6
    cnti = 12
    
    def __init__(self,gpib,refhp,name):
        self.gpib = gpib
        self.inst = Gpib.Gpib(0,self.gpib, timeout=60) # 3458A GPIB Address = self.gpib
        self.refhp = refhp
        self.name = name
        self.init_inst()

    def init_inst(self):
        # Setup HP 3458A
        self.inst.write ("PRESET NORM")
        self.inst.write ("OFORMAT ASCII")
        self.inst.write ("FUNC DCV,AUTO")
        self.inst.write ("TARM HOLD")
        self.inst.write ("TRIG LINE")
        self.inst.write ("NPLC 50")
        self.inst.write ("AZERO ON")
        self.inst.write ("NRDGS 1,AUTO")
        self.inst.write ("MEM OFF")
        self.inst.write ("END ALWAYS")
        self.inst.write ("NDIG 9")

    def switch_dci(self):
        self.inst.write ("PRESET NORM")
        self.inst.write ("NPLC 50")
        self.inst.write ("AZERO ON")
        self.inst.write ("NRDGS 1,AUTO")
        self.inst.write ("OFORMAT ASCII")
        self.inst.write ("FUNC DCI,AUTO")
        self.inst.write ("NDIG 9")
        
    def read_data(self,cmd):
        data_float = 0.0
        data_str = ""
        self.inst.write(cmd)
        
        try:
            with Timeout(20):
                data_str = self.inst.read()
        except Timeout.Timeout:
            print ("Timeout exception from dmm %s on read_data() inst.read()\n" % self.name)
            return (0,float(0))
        #print ("Reading from dmm %s = %s" % (self.name,data_str))
        try:
            data_float = float(data_str)
        except ValueError:
            print("Exception thrown by dmm %s on read_data() - ValueError = %s\n" % (self.name,data_str))
            return (0,float(0)) # Exception on float conversion, 0 = error
        return (1,data_float) # Good read, 1 = converted to float w/o exception
        
    def get_temp(self):
        global dmm_temp
        print("Reading DMM temp")
	self.inst.write("NPLC 10")
        self.inst.write("TARM SGL,1")
        self.temp_status_flag,temp = self.read_data("TEMP?")
        if (self.temp_status_flag):
            self.temp = temp
        dmm_temp = 37.5
        return self.temp
        
    def get_temp_status(self):
        return self.temp_status_flag

    def get_data(self):
        global dmm_val
        self.status_flag,data = self.read_data("TARM SGL,1")
        if (self.status_flag):
            self.data = data
            dmm_val = float(data)
            self.ppm = ((float(self.data) / self.refhp)-1)*1E6
        return self.data
        
    def exec_acal(self):
        sys.stdout.write ("\033[1;35mACAL ALL procedure start, please wait 12 minutes ")
        self.inst.write("ACAL ALL")
        for cnt in range(0 ,48):        # wait 720 seconds, print dot every 15s
            sys.stdout.write ("\033[0;43m*")
            time.sleep(15)
        print("\r\nACAL procedure done...\033[1;39m")
    
    def exec_idn(self):
        self.inst.write ("END ALWAYS")
        self.inst.write ("ID?")
        dat = self.inst.read()
        tstr = dat.split()
        if (tstr[0] == "HP3458A"):
            sys.stdout.write ("\r\n\033[1;32m%s detected...\033[1;39m" % tstr[0])
        else:
            sys.stdout.write ("\r\n\033[1;31mNo DMM present, exiting!\033[1;39m")
            quit()

    def get_data_status(self):
        return self.status_flag

with open('callog_ver.txt','wb') as b:
    kei2304 = Gpib.Gpib(0,5)               # Keithley Model 2400 GPIB address init 5
    kei2304.clear()

    cal_dcv_set = [0, 0.1, 0.5, 1, 1.5, 0, 2, 0, -20, 0, 20, 0, -200, 0, 200, 0]
    cal_dcv_rng = [0.2, 0.2, 0.2, 0.2, 2, 2, 2, 2, 20, 20, 20, 20, 200, 200, 200, 200]
    cal_dcv_str = ["-0.2 m"," -Zero"," 0.2 m"," +Zero","   -2 "," -Zero","    2 "," -Zero","  -20 "," +Zero","   20 "," -Zero"," -200 "," +Zero","  200 "," +Zero"]
    cal_dci_set = [-1E-6, 0, 1E-6, 0, -10E-6, 0, 10E-6, 0, -100E-6, 0, 100E-6, 0, -1E-3, 0, 1E-3, 0, -10E-3, 0, 10E-3, 0, -100E-3, 0, 100E-3, 0, -1, 0, 1, 0]
    cal_dci_rng = [1E-6, 1E-6, 1E-6, 1E-6, 10E-6, 10E-6, 10E-6, 10E-6, 100E-6, 100E-6, 100E-6, 100E-6, 1E-3, 1E-3, 1E-3, 1E-3, 10E-3, 10E-3, 10E-3, 10E-3, 100E-3, 100E-3, 100E-3, 100E-3, 1, 1, 1, 1]
    cal_dci_str = ["-1 u","-Zero", " 1 u", "+Zero", " -10 u", "-Zero", " 10 u", "+Zero", "-100 u", "-Zero", " 100 u", "+Zero", "-1 m", "-Zero", " 1 m", "+Zero", " -10 m", "-Zero", " 10 m", "+Zero", "-100 m", "-Zero", " 100 m", "+Zero", " -1 ", "-Zero", " 1 ", "+Zero"]
    test_dcv_set = [0, -0.002, 0.002, -0.02, 0.02, -0.2, 0.2, -1, 1, -2, 2, -10, 10, -15, 15, -20, 20, -100, 100, -200, 200, 0]
    test_dcv_rng = [0.2,  0.2,   0.2,   0.2,  0.2,  0.2, 0.2,  2, 2, 20,20,  20, 20,  20, 20,  20, 20, 200, 200, 200, 200, 20]
    test_dci_set = [0, -1E-9, 1E-9, -10E-9, 10E-9, -100E-9, 100E-9, -1E-6, 1E-6, -10E-6, 10E-6, -100E-6, 100E-6, -1E-3, 1E-3, -4E-3, 4E-3, -10E-3, 10E-3, -20E-3, 20E-3, -100E-3, 100E-3, -500E-3, 500E-3, 1, -1, 0]
    test_dci_rng = [1E-6, 1E-6, 1E-6, 1E-6, 1E-6, 1E-6,  1E-6, 1E-6, 1E-6, 10E-6, 10E-6, 100E-6, 100E-6, 1E-3, 1E-3, 10E-3, 10E-3, 10E-3, 10E-3, 100E-3, 100E-3, 100E-3, 100E-3, 1, 1, 1, 1, 1E-3]
    test_dci_str = ["Zero  ","  -1 n", "   1 n", " -10 n", "  10 n", "-100 n", " 100 n", "  -1 u", "   1 u", " -10 u", "  10 u", "-100 u", " 100 u", "  -1 m", "   1 m", "  -4 m", "   4 m", " -10 m", "  10 m", " -20 m", "  20 m", "-100 m", " 100 m", "-500 m", " 500 m", "  -1  ", "   1  ", "Zero  "]
    kei2304_idn = "KEITHLEY INSTRUMENTS INC."
    kei2304_mdl = "MODEL 2304A"
    orig_val = []
    tval = 0.0
    res4 = 3.99895
    res4k = 4000.000
    perf_dcv_start = 0.00
    perf_dcv_end = 20.00
    perf_dcv_step = 0.1
    kei_error = 0
    plc = 60 #60 Hz
    cnt = 0
    samples = 0
    nplc = 0
    val = 3.423
    meas = 0.0
    temp = 36.6
    
    for cnt in range(30):
        orig_val.append(0)
        
    sys.stdout.write("\033[0;36m Keithley 2304A PSU calibration tool \r\n  Using NI GPIB adapter with next instruments config: \r\n * GPIB 23 : Keithley 2304A PSU \r\n * GPIB 3 : HP 3458A 8.5-digit DMM \r\n ! Do not swap terminals during calibration!\033[0;39m\r\n")
    
    kei2304.write ("*IDN?")
    idn_str = kei2304.read()
    idmfg = idn_str.split(",")
    if (kei2304_idn == idmfg[0]):
        if (kei2304_mdl == idmfg[1]):
            sys.stdout.write("\033[1;31mKeithley \033[0;32mModel 2304A - detected, S/N %s, Version: %s\033[0;39m\r\n" % (idmfg[2], idmfg[3]))
            b.write("\r\nKeithley Model 2304A - detected, S/N %s, Version: %s\r\n" % (idmfg[2], idmfg[3]))
        else:
            print ("\033[0;31mIncorrect Model! Check GPIB address. Calibration cancelled.\033[0;39m")    
            quit()
    else:
        sys.stdout.write ("\033[0;31mNo Keithley instrument detected. Check GPIB address. Calibration cancelled.\033[0;39m")
        quit()
    
    # Keithley Model 2400 detected OK, proceed with reading CAL data
    
    kei2304.write ("*CLS")
    kei2304.write ("*ESE 1;*SRE 32")
    kei2304.write ("*RST")
    kei2304.write (":SENS:FUNC:CONC OFF")
    kei2304.write (":SENS:FUNC 'VOLT:DC'")
    kei2304.write (":SOUR:FUNC VOLT")
    kei2304.write (":OUTP:STAT OFF")

    print "Performance verification for Model 2304 - Voltage source"
    b.write("Pre-cal K2304 performance verification data - voltage source \r\n")

    dmm = dmm(3,dmm_val,"3458A") # GPIB 3
    dmm.exec_idn()
    dmm.get_temp()
    print ("\033[0;33mHP3458A TEMP = %2.1f C\033[0;33m" % dmm_temp)
    b.write ("HP 3458A detected\r\n TEMP? = %2.1f C\r\n" % dmm_temp)
    #print ("\033[1;31m Connect DMM volts input to SOURCE and DVM IN terminals. \033[1;39m")
    raw_input("\033[1;33mPress Enter to continue with test...\033[1;39m")
    
    #dmm.exec_acal()         # Execute ACAL on HP 3458A for best accuracy. Takes 12 minutes to complete.
    
    #Ramp 0 to 20V in 0.1 steps.
    #Measure each point
    #Report deviation
    
#    kei2304.write (":OUTP:STAT ON")
#    for cnt in range (0, 200):
        
#        tval = float(cnt) / 10.0
#        print ("Setting PSU output %5.3f" % tval)
#        kei2304.write(":SOUR:VOLT %5.3f" % tval)
        
#        val = 0 
#        time.sleep(0.1)
#        for samples in range (0, 3):
#            dmm.get_data()
#            print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
#            val += dmm_val
#        val = val / 3
#        ppmd = ((val / (tval + 0.0000000001)) - 1) * 1E6
    
#        kei2304.write(":MEAS:VOLT?")
#        psuval = float(kei2304.read())
#        kei2304.write(":MEAS:VOLT?")
#        psuval = float(kei2304.read())
    
#	ppmv = ((psuval / (val + 0.0000000001)) - 1) * 1E6

#        print ("Measured Vout = %f VDC, Deviation source %f ppm, PSU fetch: %f VDC, Delta-DMM %f ppm" % (val, ppmd, psuval, ppmv))
#        b.write ("Measured Vout = %f VDC, Deviation source %f ppm, PSU fetch: %f VDC, Delta-DMM %f ppm\r\n" % (val, ppmd, psuval, ppmv))
#    kei2304.write (":OUTP:STAT OFF")
    
#    print "Performance verification for Model 2304 - Current compliance 5A"
#    b.write("Pre-cal K2304 performance verification data - current compliance 5A \r\n")
#    print ("Connect 4OHM resistor to PSU source and sense HI/LO and DMM voltage HI/LO")
    res4 = float(raw_input("\033[1;33mEnter value of 4R resistor, in ohms, e.g. 3.99895 and press Enter...\033[1;39m"))
    print ("Resistance entered = %f Ohm" % res4)
    b.write ("Resistance 4R entered = %f Ohm\r\n" % res4)
    res4k = float(raw_input("\033[1;33mEnter value of 4KR resistor, in ohms, e.g. 3992.3 and press Enter...\033[1;39m"))
    print ("Resistance entered = %f Ohm" % res4k)
    b.write ("Resistance 4K entered = %f Ohm\r\n" % res4k)


#    raw_input("\033[1;33mPress Enter to continue with test...\033[1;39m")

#    kei2304.write (":OUTP:STAT ON")
#    for cnt in range (1, 6):
        
#        tval = float(cnt) + 0.00000001
#	kei2304.write (":OUTP:STAT ON")
#        print ("Setting PSU output %5.3f" % tval)
#        kei2304.write(":SENS:CURR:RANG MAX")
#        kei2304.write(":SOUR:VOLT 20")
#	kei2304.write(":SOUR:CURR:LIM:VAL %5.3f" % tval)
        
#        val = 0 
#        time.sleep(0.1)
#        for samples in range (0, 3):
#            dmm.get_data()
#            print ("Measured DCI %d : %.9E VDC" % (samples, dmm_val))
#            val += dmm_val
#        val = val / 3
#        val = val / res4                          # convert to current
#        ppmd = ((val / float(tval) + 0.0000001) - 1) * 1E6
        
#        kei2304.write(":MEAS:CURR?")
#        psuval = float(kei2304.read())
#        kei2304.write(":MEAS:CURR?")
#        psuval = float(kei2304.read())

#        ppmv = ((psuval / val + 0.000001) - 1) * 1E6

#        kei2304.write (":OUTP:STAT OFF")
#        print ("Sleeping 10 seconds to allow resistors to cool")
#        time.sleep(10)                            
       
#        print ("Measured Iout = %f DCI, Deviation source %f ppm, PSU fetch: %f ADC, Delta %f ppm" % (val, ppmd, psuval, ppmv))
#        b.write ("Measured Iout = %f DCI, Deviation source %f ppm, PSU fetch: %f ADC, Delta %f ppm\r\n" % (val, ppmd, psuval, ppmv))
#    kei2304.write (":OUTP:STAT OFF")

#    kei2304.write (":OUTP:STAT OFF")
#    raw_input("\033[1;33mPress Enter to continue with test...\033[1;39m")

#    print "Performance verification for Model 2304 DVM IN"
#    b.write("Pre-cal K2304 performance verification data - DVM IN \r\n")

#    kei2304.write (":OUTP:STAT ON")
#    kei2304.write(":SENS:FUNC DVM")
#    for cnt in range (-30, 200):
        
#        tval = float(cnt) / 10.0
#        print ("Setting PSU output %5.3f" % tval)
#        kei2304.write(":SOUR:VOLT %5.3f" % tval)
        
#        val = 0 
#        time.sleep(0.1)
#        for samples in range (0, 3):
#            dmm.get_data()
#            print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
#            val += dmm_val
#        val = val / 3
#        ppmd = ((val / (tval + 0.0000000001)) - 1) * 1E6
    
#        kei2304.write(":MEAS:DVM?")
#        psuval = float(kei2304.read())
#        kei2304.write(":MEAS:DVM?")
#        psuval = float(kei2304.read())
    
#	ppmv = ((psuval / (val + 0.0000000001)) - 1) * 1E6

#        print ("Measured Vout = %f VDC, Deviation source %f ppm, PSU fetch: %f VDC, Delta-DMM %f ppm" % (val, ppmd, psuval, ppmv))
#        b.write ("Measured Vout = %f VDC, Deviation source %f ppm, PSU fetch: %f VDC, Delta-DMM %f ppm\r\n" % (val, ppmd, psuval, ppmv))
    kei2304.write (":OUTP:STAT OFF")

    raw_input("\033[1;33mPress Enter to start cal...\033[1;39m")
    print "Calibration for Model 2304"
    b.write("Calibration K2304 \r\n")

    kei2304.write (":CAL:PROT:CODE 'KI002304'")
    kei2304.write (":CAL:PROT:INIT")
    print ("Step 0, 19V output")
    kei2304.write (":CAL:PROT:STEP0 19")

    val = 0 
    time.sleep(0.1)
    for samples in range (0, 10):
        dmm.get_data()
        print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
        val += dmm_val
    val = val / 10
    ppmd = (((val / 19) - 1) * 1E6)

    print ("Measured Vout = %f VDC, Deviation source %f ppm" % (val, ppmd))
    b.write ("Measured Vout = %f VDC, Deviation source %f ppm\r\n" % (val, ppmd))
    kei2304.write (":CAL:PROT:STEP1 %8.6f" % val)

    print ("Step 2, 19V DVM output")

    val = 0 
    time.sleep(0.1)
    for samples in range (0, 10):
        dmm.get_data()
        print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
        val += dmm_val
    val = val / 10
    ppmd = (((val / 19) - 1) * 1E6)

    print ("Measured Vout = %f VDC, Deviation source %f ppm" % (val, ppmd))
    b.write ("Measured Vout = %f VDC, Deviation source %f ppm\r\n" % (val, ppmd))
    kei2304.write (":CAL:PROT:STEP2 %8.6f" % val)

    kei2304.write (":CAL:PROT:STEP3")
    print("STEP 3, connect 4 ohm")
    raw_input("\033[1;33mPress Enter to continue with cal...\033[1;39m")

    kei2304.write (":CAL:PROT:STEP4 1.9")

    print ("Step 4, 1.9A over 4R")

    val = 0 
    time.sleep(0.1)
    for samples in range (0, 5):
        dmm.get_data()
        print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
        val += dmm_val
    val = val / 5
    ccurr = val / res4
    ppmd = (((val / 1.9) - 1) * 1E6)

    print ("Measured Vout = %f VDC, Current = %f A, Deviation source %f ppm" % (val, ccurr, ppmd))
    b.write ("Measured Vout = %f VDC, Current = %f A, Deviation source %f ppm\r\n" % (val, ccurr, ppmd))
    kei2304.write (":CAL:PROT:STEP5 %8.6f" % ccurr)

    print ("Step 5, 1.9A over 4R")

    val = 0 
    time.sleep(0.1)
    for samples in range (0, 5):
        dmm.get_data()
        print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
        val += dmm_val
    val = val / 5
    ccurr = val / res4
    ppmd = (((val / 1.9) - 1) * 1E6)

    print ("Measured Vout = %f VDC, Current = %f A, Deviation source %f ppm" % (val, ccurr, ppmd))
    b.write ("Measured Vout = %f VDC, Current = %f A, Deviation source %f ppm\r\n" % (val, ccurr, ppmd))
    kei2304.write (":CAL:PROT:STEP6 %8.6f" % ccurr)

    print("STEP 3, connect 4000 ohm")
    raw_input("\033[1;33mPress Enter to continue with cal...\033[1;39m")
    print ("Step 7, 5mA over 4K")
    kei2304.write (":CAL:PROT:STEP7")

    val = 0 
    time.sleep(0.1)
    for samples in range (0, 5):
        dmm.get_data()
        print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
        val += dmm_val
    val = val / 5
    ccurr = val / res4k
    ppmd = (((val / 0.005) - 1) * 1E6)

    print ("Measured Vout = %f VDC, Current = %f A, Deviation source %f ppm" % (val, ccurr, ppmd))
    b.write ("Measured Vout = %f VDC, Current = %f A, Deviation source %f ppm\r\n" % (val, ccurr, ppmd))
    kei2304.write (":CAL:PROT:STEP8 %8.6E" % ccurr)

    raw_input("\033[1;33mPress Enter to save cal...\033[1;39m")
    print ("Saving calibration with date 2016 04 11")
    b.write ("Saving calibration with date 2016 04 11\r\n")
    
    kei2304.write (":CAL:PROT:DATE 2016,04,11")
    kei2304.write (":CAL:PROT:SAVE")
    kei2304.write (":CAL:PROT:LOCK")

    quit()
    
    kei2304.write ("*RST")
    kei2304.write ("*CLS")                                  # Init 2304A
    kei2304.write ("*ESE1;*SRE 32")
    kei2304.write (":CAL:PROT:CODE 'KI002304'")             # Unlock CAL mode
    kei2304.write (":CAL:PROT:INIT")                        # Init calibration 
    
    sys.stdout.write("Starting DCV calibration...\r\n\r\n")
    
    for cnt in range (0, len(cal_dcv_set)):
        #sys.stdout.write(" Calibration step %d " % cnt)
        
        kei2304.write (":SOUR:VOLT:RANGE %3.1f" % cal_dcv_rng[cnt])
        kei2304.write (":SOUR:VOLT %3.1f" % cal_dcv_set[cnt])
        
        val = 0 
        time.sleep(1)
        for samples in range (0, 10):
            dmm.get_data()
            print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
            val += dmm_val
        val = val / 10
        
        orig_val[cnt] = val
        sys.stdout.write("DCV Calibration step %sV : %.9E VDC" % (cal_dcv_str[cnt], val))
        
        kei2304.write (":CAL:PROT:SOUR %3.8f;*OPC" % val)
        time.sleep(1)
        kei2304.write (":SYST:ERR?")
        kei_error = kei2304.read().split(",")
        if (int(kei_error[0]) != 0):
            print "\033[1;31mKeithley 2400 error : %d\033[1;39m" % int(kei_error[0])
            quit()
        #kei2304.write ("*STB?")
        #kei2304.write ("*ESR?")
        
        kei2304.write (":CAL:PROT:SENS %3.8f;*OPC" % val)
        time.sleep(1)
        kei2304.write (":SYST:ERR?")
        kei_error = kei2304.read().split(",")
        if (int(kei_error[0]) != 0):
            print "\033[1;31mKeithley 2400 error : %d\033[1;39m" % int(kei_error[0])
            quit()
        #kei2304.write ("*STB?")
        #kei2304.write ("*ESR?")
        
        if (cnt % 2 == 0):
            ppms = ((val / (cal_dcv_set[cnt] + 1E-16)) - 1) * 1E6
            sys.stdout.write("\033[1;35m Delta: %.3f ppm\033[1;39m\r\n" % ppms)
            b.write("DCV Calibration step %sV , measured : %.9E VDC, deviation %.3f ppm\r\n" % (cal_dcv_str[cnt], val, ppms))
        else:
            sys.stdout.write("\r\n")
        cnt = cnt + 1
  
    sys.stdout.write("\r\n")
    sys.stdout.write("\033[0;31;43mConnect DMM volts input and 4 OHM resistor to SOURCE and SENSE.\033[0;39;49m\r\n")
    
    sys.stdout.write("\r\n")
    sys.stdout.write("\033[0;31;43mConnect DMM volts input and 4000 OHM resistor to SOURCE and SENSE.\033[0;39;49m\r\n")
    
    raw_input("Press Enter to continue...")
    
    sys.stdout.write("Starting DCI calibration...\r\n\r\n")
    
    dmm.switch_dci()

    kei2304.write (":SOUR:FUNC CURR")
    kei2304.write (":SENS:VOLT:PROT 20")
    kei2304.write (":SENS:VOLT:RANG 20")
    kei2304.write (":OUTP:STAT ON")
    
    for cnt in range (0, len(cal_dci_set)):
        
        kei2304.write (":SOUR:CURR:RANGE %3.1e" % cal_dci_rng[cnt])
        kei2304.write (":SOUR:CURR %3.1e" % cal_dci_set[cnt])
        
        val = 0 
        time.sleep(1)
        for samples in range (0, 10):
            dmm.get_data()
            print ("Measured DCI %d : %.9E ADC" % (samples, dmm_val))
            val += dmm_val
        val = val / 10

        sys.stdout.write("DCI Calibration step %sA : %.9E ADC" % (cal_dci_str[cnt], val))
        
        kei2304.write (":CAL:PROT:SOUR %3.8e;*OPC" % val)
        time.sleep(1)
        kei2304.write (":SYST:ERR?")
        kei_error = kei2304.read().split(",")
        if (int(kei_error[0]) != 0):
            print "\033[1;31mKeithley 2400 error : %d\033[1;39m" % int(kei_error[0])
            quit()
        #kei2304.write ("*STB?")
        #kei2304.write ("*ESR?")
        
        kei2304.write (":CAL:PROT:SENS %3.8e;*OPC" % val)
        time.sleep(1)
        kei2304.write (":SYST:ERR?")
        kei_error = kei2304.read().split(",")
        if (int(kei_error[0]) != 0):
            print "\033[1;31mKeithley 2400 error : %d\033[1;39m" % int(kei_error[0])
            quit()
        #kei2304.write ("*STB?")
        #kei2304.write ("*ESR?")
        
        if (cnt % 2 == 0):
            ppms = ((val / (cal_dci_set[cnt] + 1E-16)) - 1) * 1E6
            sys.stdout.write("\033[1;36m Delta: %.3f ppm\033[1;39m\r\n" % ppms)
            b.write("DCI Calibration step %sA , measured : %.9E ADC, deviation %.3f ppm\r\n" % (cal_dci_str[cnt], val, ppms))
        else:
            sys.stdout.write("\r\n")
        cnt = cnt + 1
    
    kei2304.write (":CAL:PROT:DATE 2016, 03, 22")
    kei2304.write (":CAL:PROT:SAVE")
    kei2304.write (":CAL:PROT:LOCK")
    kei2304.write (":OUTP:STAT OFF")

    print("\033[0;30;42mCalibration complete! www.xDevs.com \033[0;39;49m")
    kei2304.write (":SYST:BEEP:IMM 100, 1")
    kei2304.write (":SYST:BEEP:IMM 300, 1")
    kei2304.write (":SYST:BEEP:IMM 500, 1")
    kei2304.write (":SYST:BEEP:IMM 900, 1")
    kei2304.write (":SYST:BEEP:IMM 1200, 1")
    kei2304.write (":SYST:BEEP:IMM 1500, 1")
    kei2304.write (":SYST:BEEP:IMM 2000, 1")
    
    sys.stdout.write("\r\n\r\nPerformance DCI test\r\n")
    dmm.init_inst()
    dmm.get_temp()
    dmm.switch_dci()
    sys.stdout.write("3458A TEMP? = %02.1f \r\n" % (dmm_temp))
    
    for cnt in range (0, len(test_dci_set)):
        kei2304.write (":SOUR:FUNC CURR")
        kei2304.write (":SENS:VOLT:PROT 20")
        kei2304.write (":SENS:VOLT:RANG 20")
        kei2304.write (":OUTP:STAT ON")
        kei2304.write (":SOUR:CURR:RANGE %3.1e" % test_dci_rng[cnt])
        kei2304.write (":SOUR:CURR %3.1e" % test_dci_set[cnt])
        time.sleep(1)

        val = 0
        for samples in range (0, 5):
            dmm.get_data()
            print ("Test DCI %d : %.9E ADC" % (samples, dmm_val))
            val += dmm_val
        val = val / 5
        ppm = (((val+1e-16) / (test_dci_set[cnt] + 1E-16)) - 1) * 1E6
        sys.stdout.write(" Verification step %sA : %.9E ADC [deviation %6.2f ppm]\r\n" % (test_dci_str[cnt], val, ppm))
        b.write("DCI Verification step %sA , measured : %.9E ADC, deviation %.3f ppm\r\n" % (test_dci_str[cnt], val, ppm))
        cnt = cnt + 1

    kei2304.write (":OUTP:STAT OFF")
    sys.stdout.write("\033[0;31;43mConnect DMM VOLTAGE input to 2400 INPUT/OUTPUT jacks.\033[0;39;49m\r\n\r\n")
    print ":SYST:BEEP:IMM 1000, 5"
    raw_input("Press Enter to continue...")
        
    sys.stdout.write("Performance DCV test\r\n")
    dmm.init_inst()
    dmm.get_temp()
    print ("\033[0;33mHP3458A TEMP = %2.1f C\033[0;33m" % dmm_temp)
    
    for cnt in range (0, len(test_dcv_set)):
        kei2304.write (":SOUR:FUNC VOLT")
        kei2304.write (":SENS:CURR:PROT 0.1")
        kei2304.write (":SENS:CURR:RANG 0.1")
        kei2304.write (":OUTP:STAT ON")
        time.sleep(1)
        kei2304.write (":SOUR:VOLT:RANGE %3.1e" % test_dcv_rng[cnt])
        kei2304.write (":SOUR:VOLT %3.1e" % test_dcv_set[cnt])
        time.sleep(1)

        val = 0
        for samples in range (0, 5):
            dmm.get_data()
            print ("Test DCV %d : %.9E VDC" % (samples, dmm_val))
            val += dmm_val
        val = val / 5
        ppm = (((val+1e-16) / (test_dcv_set[cnt] + 1E-16)) - 1) * 1E6
        
        sys.stdout.write(" Verification step  %02d (%.9E) : %.9E VDC [deviation %6.2f ppm]\r\n" % (cnt, test_dcv_set[cnt], val, ppm))
        b.write("DCV Verification step %02d (%.9E), measured : %.9E VDC, deviation %.3f ppm\r\n" % (cnt, test_dcv_set[cnt], val, ppm))
        cnt = cnt + 1

    # Post-cal constants 
    print "Reading Post-cal calibration data from Model 2400"
    b.write("Post-cal K2400 Calibration data\r\n")
    
    for cnt in range (0, 4):
        kei2304.write (":SOUR:VOLT:RANGE %3.1f" % cal_dcv_rng[cnt*4])
        kei2304.write (":CAL:PROT:SENS:DATA?")
        kdata = kei2304.read()
        sys.stdout.write ("%s" % kdata)
        b.write("%3.1f DCV Range, SENS:DATA? = %s\r\n" % (cal_dcv_rng[cnt*4], kdata))
        kei2304.write (":CAL:PROT:SOUR:DATA?")
        kdata = kei2304.read()
        sys.stdout.write ("%s" % kdata)
        b.write("%3.1f DCV Range, SOUR:DATA? = %s\r\n" % (cal_dcv_rng[cnt*4], kdata))
        cnt = cnt + 1
    
    kei2304.write (":CAL:PROT:LOCK")
    kei2304.write (":SENS:FUNC 'CURR:DC'")
    kei2304.write (":SOUR:FUNC CURR")
    kei2304.write (":CAL:PROT:CODE 'KI002400'")
    
    for cnt in range (0, 7):
        kei2304.write (":SOUR:CURR:RANGE %3.1e" % cal_dci_rng[cnt*4])
        kei2304.write (":CAL:PROT:SENS:DATA?")
        kdata = kei2304.read()
        sys.stdout.write ("%s" % kdata)
        b.write("%3.1e DCI Range, SENS:DATA? = %s\r\n" % (cal_dci_rng[cnt*4], kdata))
        kei2304.write (":CAL:PROT:SOUR:DATA?")
        kdata = kei2304.read()
        b.write("%3.1e DCI Range, SOUR:DATA? = %s\r\n" % (cal_dci_rng[cnt*4], kdata))
        sys.stdout.write ("%s" % kdata)
        cnt = cnt + 1

    kei2304.write (":CAL:PROT:LOCK")
    b.write ("Calibration complete!")
    b.close()